/*
 * pngfprint: calculate a representation-independent checksum for PNG image.
 *
 * written by Alexander Lehmann <alex@hal.rhein-main.de>
 *
 * v0.1 08.06.96
 *
 */

#include <stdio.h>
#include <stdlib.h>

#ifdef NeXT
#include <libc.h>
#else
#include <unistd.h>
#endif

#define PNG_INTERNAL

#include "png.h"
#include "md5.h"

#define CHUNKNAME "fiNG"

typedef unsigned char byte;

/* map file into memory, if not supported, just malloc enough space and read
   it */
/* (well, not yet) */

byte *data;
size_t size;
byte *current;

byte *mapfile(FILE *fp)
{
  fseek(fp, 0, SEEK_END);
  size=ftell(fp);
  fseek(fp, 0, SEEK_SET);
  if((data=malloc(size))==NULL) {
    fclose(fp);
    fprintf(stderr, "out of memory\n");
    return NULL;
  }
  fread(data, 1, size, fp);
  fclose(fp);
  current=data;
  return data;
}

void unmapfile(void)
{
  free(data);
}

void read_mem_data(png_structp png_ptr, png_bytep data, png_uint_32 length)
{
  if(current+length<=data+size) {
    memcpy(data, current, length);
    current+=length;
  } else
    png_error(png_ptr, "premature end of input file");
}

struct MD5Context md5ctxt;
unsigned char result[16];

void init_calc(void)
{
  MD5Init(&md5ctxt);
}

void finish_calc(void)
{
  MD5Final(result, &md5ctxt);
}

void calc(png_byte *p, png_uint_32 len)
{
  MD5Update(&md5ctxt, p, len);
}

void calc0(png_uint_32 len)
{
  while(len--) {
    MD5Update(&md5ctxt, "", 1);
  }
}

void calculate_24(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 1);
    calc(image++, 1);
    calc(image, 1);
    calc(image++, 1);
    calc(image, 1);
    calc(image++, 1);
    calc0(2);
  }
}

void calculate_48(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 6);
    image+=6;
    calc0(2);
  }
}

void calculate_24a(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 1);
    calc(image++, 1);
    calc(image, 1);
    calc(image++, 1);
    calc(image, 1);
    calc(image++, 1);
    calc(image, 1);
    calc(image++, 1);
  }
}

void calculate_48a(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 8);
    image+=8;
  }
}

void calculate_g8(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 1);
    calc(image, 1);
    calc(image, 1);
    calc(image, 1);
    calc(image, 1);
    calc(image++, 1);
    calc0(2);
  }
}

void calculate_g16(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 2);
    calc(image, 2);
    calc(image, 2);
    image+=2;
    calc0(2);
  }
}

void calculate_g8a(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 1);
    calc(image, 1);
    calc(image, 1);
    calc(image, 1);
    calc(image, 1);
    calc(image++, 1);
    calc(image, 1);
    calc(image++, 1);
  }
}

void calculate_g16a(png_byte *image, png_uint_32 len)
{
  while(len--) {
    calc(image, 2);
    calc(image, 2);
    calc(image, 2);
    image+=2;
    calc(image, 2);
    image+=2;
  }
}

void calculate_pal(png_byte *image, png_colorp palette, png_uint_32 len)
{
  while(len--) {
    calc(&palette[*image].red,1);
    calc(&palette[*image].red,1);
    calc(&palette[*image].green,1);
    calc(&palette[*image].green,1);
    calc(&palette[*image].blue,1);
    calc(&palette[*image++].blue,1);
    calc0(2);
  }
}

int processfile(char *name)
{
  png_structp png_ptr;
  png_infop info_ptr;
  int i;
  png_bytep image,p;
  png_bytepp rowpointers;

  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
				   NULL, NULL, NULL);
  if (!png_ptr)
    return 1;
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
    return 1;
  }
  if (setjmp(png_ptr->jmpbuf)) {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    return 1;
  }

  png_set_read_fn(png_ptr, NULL, read_mem_data);

  png_read_info(png_ptr, info_ptr);

  if(info_ptr->bit_depth<8)
    png_set_packing(png_ptr);
  
  png_read_update_info(png_ptr, info_ptr);

  if((rowpointers=malloc(info_ptr->height*sizeof(char *)))==NULL) {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    return 0;
  }

  if((image=malloc(info_ptr->height*info_ptr->rowbytes))==NULL) {
    free(rowpointers);
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    return 0;
  }

  p=image;
  for(i=0;i<info_ptr->height;i++) {
    rowpointers[i]=p;
    p+=info_ptr->rowbytes;
  }

  png_read_image(png_ptr, rowpointers);

  png_read_end(png_ptr, NULL);

  printf("%ldx%ld %ld\n", info_ptr->width, info_ptr->height, info_ptr->rowbytes);

  init_calc();

  switch(info_ptr->color_type) {
    case PNG_COLOR_TYPE_RGB:
      if(info_ptr->bit_depth==8) {
	calculate_24(image, info_ptr->width*info_ptr->height);
      } else {
	calculate_48(image, info_ptr->width*info_ptr->height);
      }
      break;

    case PNG_COLOR_TYPE_RGB_ALPHA:
      if(info_ptr->bit_depth==8) {
	calculate_24a(image, info_ptr->width*info_ptr->height);
      } else {
	calculate_48a(image, info_ptr->width*info_ptr->height);
      }
      break;

    case PNG_COLOR_TYPE_GRAY:
      if(info_ptr->bit_depth==16) {
	calculate_g16(image, info_ptr->width*info_ptr->height);
      } else {
	calculate_g8(image, info_ptr->width*info_ptr->height);
      }
      break;

    case PNG_COLOR_TYPE_GRAY_ALPHA:
      if(info_ptr->bit_depth==16) {
	calculate_g16a(image, info_ptr->width*info_ptr->height);
      } else {
	calculate_g8a(image, info_ptr->width*info_ptr->height);
      }
      break;

    case PNG_COLOR_TYPE_PALETTE:
      calculate_pal(image, info_ptr->palette, info_ptr->width*info_ptr->height);
      break;

    default:
      fprintf(stderr, "unhandled colortype %d\n", info_ptr->color_type);
      return 1;
  }

  finish_calc();

  for(i=0;i<16;i++)
    printf("%02x", result[i]);
  printf(" %s\n", name);

  png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

  free(rowpointers);
  free(image);

  return 0;
}

int findprint(char *name)
{
  png_byte *p=data+8;
  png_uint_32 len;
  int i;

  while(p<data+size-8) {
    if(memcmp(p, "\0\0\0\x10" CHUNKNAME, 8)==0) {
      for(i=0;i<16;i++)
        printf("%02x", p[i+8]);
      printf(" %s\n", name);
      return 1;
    }
    if(memcmp(p, "\0\0\0\0IEND", 8)==0) {
      return 0;
    }
    len=(p[0]<<24)|(p[1]<<16)|(p[2]<<8)|p[3];
    p+=len+12;
  }
  return 0;
}

void fwrite_crc(void *data, size_t s, size_t n, FILE *fp, png_struct *crc)
{
  fwrite(data, s, n, fp);
  png_calculate_crc(crc, data, s*n);
}

int doadd(char *name)
{
  char name2[1000];
  FILE *fp;
  png_uint_32 len;
  png_struct *png_ptr=png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

  if(png_ptr==NULL) {
    fprintf(stderr, "out of memory\n");
    return 1;
  }

  strcpy(name2, name);
  strcat(name2, ".bak");

  unlink(name2);

  if(rename(name, name2)) {
    perror(name2);
    return 1;
  }

  if((fp=fopen(name, "rb"))!=NULL) {
    fprintf(stderr, "%s: cannot rename (probably ro filesystem or filename problem)\n", name);
    fclose(fp);
    return 1;
  }

  if((fp=fopen(name, "wb"))==NULL) {
    perror(name);
    return 1;
  }

  fwrite(data, 1, 33, fp);

  current=data+33;

  if(setjmp(png_ptr->jmpbuf)) {
    fprintf(stderr, "libpng returns error condition, recover file from .bak copy\n");
    png_destroy_write_struct(&png_ptr, NULL);
    fclose(fp);
    return 1;
  }

  png_init_io(png_ptr, fp);

  png_write_chunk(png_ptr, CHUNKNAME, result, 16);

  while(current<data+size) {
    len=(current[0]<<24)|(current[1]<<16)|(current[2]<<8)|current[3];
    if(strncmp(current+4, CHUNKNAME, 4)!=0) {
      fwrite(current, 1, len+12, fp);
    }
    current+=len+12;
  }

  png_destroy_write_struct(&png_ptr, NULL);

  fclose(fp);

  return 0;
}

int main(int argc, char *argv[])
{
  FILE *fp;
  int i;
  int addchunk=0;

  if(argc==1) {
    fprintf(stderr, "usage: %s {-a} file ...\n", argv[0]);
    return -1;
  }

  if(strcmp(argv[1], "-a")==0) {
    addchunk=1;
    argv++;
    argc--;
  }

  for(i=1;i<argc;i++) {
    if((fp=fopen(argv[i],"rb"))==NULL) {
      perror(argv[i]);
    } else {
      if(mapfile(fp)) {
        if(!findprint(argv[i])) {
	  processfile(argv[i]);
	  if(addchunk) {
	    doadd(argv[i]);
	  }
	}
      }
      fclose(fp);
    }
  }
  return 0;
}
